# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


def satellites():
    from .satellite_data import SatelliteData

    return SatelliteData().get()


def planets():
    from .planet_data import PlanetData

    return PlanetData().get()


def astronauts():
    from .astronaut_data import AstronautData

    return AstronautData().get()


def no_table():
    from .no_table_data import NoTable

    return NoTable().get()
